package edu.umich.languageid.hmm;

import java.util.ArrayList;
import java.util.List;

public class DefaultViterbiNode<T> implements ViterbiNode<T>
{
	List<ViterbiTransition<T>> previousNodeTransitions = new ArrayList<ViterbiTransition<T>>();
	
	T value;
	String label;
	
	double emissionProbability = 0.0001;
	
	double bestScore = Double.NEGATIVE_INFINITY;
	double confidence = 1;
	ViterbiNode<T> bestPreviousState = null;
	boolean isTerminal;
	
	public DefaultViterbiNode(T value, String label)
	{
		this.value = value;
		this.label = label;
	}
	
	public DefaultViterbiNode(T value, String label, boolean isTerminal)
	{
		this.value = value;
		this.label = label;
		this.isTerminal = isTerminal;
	}
	
	public void addPreviousNode(ViterbiTransition<T> transition)
	{
		previousNodeTransitions.add(transition);
		recalculateScore();
	}
	
	public void addPreviousNode(ViterbiNode<T> previousNode, double transitionProb)
	{
		previousNodeTransitions.add(new DefaultViterbiTransition<T>(previousNode, transitionProb));
		recalculateScore();
	}
	
	public void addPreviousNodes(List<ViterbiTransition<T>> transitions)
	{
		previousNodeTransitions.addAll(transitions);
		recalculateScore();
	}

	void recalculateScore()
	{		
		bestScore = Double.NEGATIVE_INFINITY;
		bestPreviousState = null;
		double totalWeight = 0;
		double bestTransitionScore = Double.NEGATIVE_INFINITY;
		for(ViterbiTransition<T> transition : previousNodeTransitions)
		{
			totalWeight += Math.exp(transition.getScore());
			if(transition.getScore() > bestTransitionScore)
				bestTransitionScore = transition.getScore();
			if(transition.getScore() + Math.log(emissionProbability) > bestScore)
			{
				bestScore = transition.getScore() + Math.log(emissionProbability);
				bestPreviousState = transition.getPreviousNode();
			}
		}
		confidence = Math.exp(bestTransitionScore)/totalWeight;
	}

	public double getScore()
	{
		return bestScore;
	}

	public ViterbiNode<T> getBestPreviousState()
	{
		return bestPreviousState;
	}

	public T getValue()
	{
		return value;
	}

	public String getLabel()
	{
		return label;
	}

	public boolean isTerminal()
	{
		return isTerminal;
	}

	public void setEmissionProbability(double emissionProbability)
	{
		this.emissionProbability = emissionProbability;
		if(emissionProbability == 0.0)
			this.emissionProbability = 0.0001;
		recalculateScore();
	}

	@Override
	public String toString()
	{
		return "DefaultViterbiNode [value=" + value + ", label=" + label + "]";
	}

	public double getPreviousNodeConfidence()
	{
		return confidence;
	}
}
