package edu.umich.languageid.hmm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import edu.umich.languageid.LabeledInstance;
import edu.umich.languageid.Classifier;
import edu.umich.languageid.hmm.ViterbiHeadNode;

public class HiddenMarkovModel<V>
{
	Collection<String> labels;
	Map<String, Map<String, Double>> transitionProbabilities = new HashMap<String, Map<String, Double>>();
	
	Classifier<V> classifier;
	
	public HiddenMarkovModel(Collection<String> labels, Classifier<V> classifier)
	{
		this.labels = labels;
		this.classifier = classifier;
		initializeTransitionProbabilities();
	}

	void initializeTransitionProbabilities()
	{
		transitionProbabilities.put("__HEAD__", new HashMap<String, Double>());
		for(String label1 : labels)
		{
			transitionProbabilities.put(label1, new HashMap<String, Double>());
			for(String label2 : labels)
			{
				transitionProbabilities.get(label1).put(label2, 1.0/labels.size());
			}
			
			// special case for head
			transitionProbabilities.get("__HEAD__").put(label1, 1.0/labels.size());
			
			// special case for tail
			transitionProbabilities.get(label1).put("__TAIL__", 1.0/labels.size());
		}
	}
	
	public List<LabeledInstance<V>> labelSequence(List<V> sequence)
	{
		ViterbiNode<V> tail = initializeSequence(sequence);
		return unwindModel(tail);
	}

	ViterbiNode<V> initializeSequence(List<V> sequence)
	{
		ViterbiNode<V> head = new ViterbiHeadNode<V>();
		List<ViterbiNode<V>> previousNodes = new ArrayList<ViterbiNode<V>>();
		previousNodes.add(head);
		
		for(V item : sequence)
		{
			List<ViterbiNode<V>> currentNodes = new ArrayList<ViterbiNode<V>>();
			
			// The classifier may not give probabilities, so we need to normalize first
			Map<String, Double> emissionProbabilities = new HashMap<String, Double>();
			double totalEmissionProbability = 0;
			
			for(String label : labels)
			{
				// set up trellis with transition probabilities
				ViterbiNode<V> node = new DefaultViterbiNode<V>(item, label);
				for(ViterbiNode<V> previousNode : previousNodes)
				{
					node.addPreviousNode(previousNode, transitionProbabilities.get(previousNode.getLabel()).get(node.getLabel()));
				}	
				
				// calculate emission probability
				double emissionProbability = classifier.probability(node.getValue(), label);
				totalEmissionProbability += emissionProbability;
				emissionProbabilities.put(label, emissionProbability);
				
				currentNodes.add(node);
			}
			
			// use the normalized versions of the emission probabilities
			for(ViterbiNode<V> node : currentNodes)
			{
				if(totalEmissionProbability > 0)
					node.setEmissionProbability(emissionProbabilities.get(node.getLabel()) / totalEmissionProbability);
				else
					node.setEmissionProbability(0.5);
			}
			
			previousNodes = currentNodes;
		}
		
		ViterbiNode<V> tail = new DefaultViterbiNode<V>(null, "__TAIL__", false);
		for(ViterbiNode<V> previousNode : previousNodes)
		{
			tail.addPreviousNode(previousNode, transitionProbabilities.get(previousNode.getLabel()).get(tail.getLabel()));
		}
		
		return tail;
	}
	
	List<LabeledInstance<V>> unwindModel(ViterbiNode<V> tail)
	{
		LinkedList<LabeledInstance<V>> finalLabeledSequence = new LinkedList<LabeledInstance<V>>();
		
		double confidence = tail.getPreviousNodeConfidence();
		
		ViterbiNode<V> currentNode = tail.getBestPreviousState();
		while(!currentNode.isTerminal())
		{
			finalLabeledSequence.addFirst(new LabeledInstance<V>(currentNode.getValue(), currentNode.getLabel(), confidence));
			
			confidence = currentNode.getPreviousNodeConfidence();
			
			currentNode = currentNode.getBestPreviousState();	
		}
		
		return new ArrayList<LabeledInstance<V>>(finalLabeledSequence);
	}

	public void updateTransitionProbabilities(Map<String, Map<String, Double>> transitionProbabilities2)
	{
		// iterate to avoid overwriting the special head and tail transition probabilities
		for(String fromLabel : transitionProbabilities2.keySet())
		{
			for(String toLabel : transitionProbabilities2.get(fromLabel).keySet())
			{
				transitionProbabilities.get(fromLabel).put(toLabel, transitionProbabilities2.get(fromLabel).get(toLabel));
			}
		}
	}
	
	public void updateClassifier(Classifier<V> newClassifier)
	{
		this.classifier = newClassifier;
	}
}
