package edu.umich.languageid.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import cc.mallet.pipe.Noop;
import cc.mallet.pipe.Pipe;
import cc.mallet.types.Instance;
import cc.mallet.types.InstanceList;
import edu.umich.languageid.FeatureConfiguration;
import edu.umich.languageid.io.WordInstanceBuilder;

public class TrainingDataReader
{
	FeatureConfiguration features;
	Pipe pipe;
	int numWordsInSample;
	
	public TrainingDataReader(int numWordsInSample)
	{
		this(new FeatureConfiguration(), numWordsInSample, new Noop());
	}
	
	public TrainingDataReader(FeatureConfiguration features, int numWordsInSample, Pipe pipe)
	{
		this.features = features;
		this.pipe = pipe;
		this.numWordsInSample = numWordsInSample;
	}
	
	public Map<String, Map<String, Integer>> labeledFeatureCounts(File lang1Sample, String lang1Label, File lang2Sample, String lang2Label)
	{
		Map<String, Map<String, Integer>> labeledFeatureCounts = new HashMap<String, Map<String, Integer>>();
		labeledFeatureCounts.put(lang1Label, featureCounts(lang1Sample));
		labeledFeatureCounts.put(lang2Label, featureCounts(lang2Sample));
		return labeledFeatureCounts;
	}
	
	public Map<String, Map<String, Integer>> labeledFeatureCounts(List<File> langSamples, List<String> langLabels)
	{
		Map<String, Map<String, Integer>> labeledFeatureCounts = new HashMap<String, Map<String, Integer>>();
		for(int k=0; k<langLabels.size(); ++k)
		{
			labeledFeatureCounts.put(langLabels.get(k), featureCounts(langSamples.get(k)));
		}
		return labeledFeatureCounts;
	}
	
	public Map<String, Map<String, Integer>> labeledFeatureCounts(Map<String, File> samplesByID)
	{
		List<String> languageIDs = new ArrayList<String>();
		List<File> languageSampleFiles = new ArrayList<File>();
		for(Map.Entry<String, File> entry : samplesByID.entrySet())
		{
			languageIDs.add(entry.getKey());
			languageSampleFiles.add(entry.getValue());
		}
		return labeledFeatureCounts(languageSampleFiles, languageIDs);
	}
	
	public int numWordsInSample()
	{
		return numWordsInSample;
	}

	public InstanceList readTrainingData(File lang1Sample, String lang1Label, File lang2Sample, String lang2Label)
	{
		String[] labels = new String[2];
		labels[0] = lang1Label;
		labels[1] = lang2Label;
		
		//Alphabet dataAlphabet = new Alphabet(String.class);
		//Alphabet targetAlphabet = new Alphabet(labels);
		InstanceList lang1InstanceList = new InstanceList(pipe);
		InstanceList lang2InstanceList = new InstanceList(pipe);
				
		for(Instance i : readLanguageTrainingData(lang1Sample, lang1Label))
		{
			lang1InstanceList.addThruPipe(i);
		}
		
		for(Instance i : readLanguageTrainingData(lang2Sample, lang2Label))
		{
			lang2InstanceList.addThruPipe(i);
		}
		
		InstanceList instanceList = new InstanceList(pipe);
		
		Random r = new Random();
		for(int k=0; k<numWordsInSample; ++k)
		{
			instanceList.add(lang1InstanceList.get(r.nextInt(lang1InstanceList.size())));
			instanceList.add(lang2InstanceList.get(r.nextInt(lang2InstanceList.size())));
		}
		
		return instanceList;
	}
	
	public InstanceList readTrainingData(List<File> langSamples, List<String> langLabels)
	{
		//Alphabet dataAlphabet = new Alphabet(String.class);
		//Alphabet targetAlphabet = new Alphabet(labels);
		
		InstanceList[] instanceLists = new InstanceList[langSamples.size()];
		
		for(int k=0; k<langSamples.size(); ++k)
		{
			instanceLists[k] = new InstanceList(pipe);
			for(Instance i : readLanguageTrainingData(langSamples.get(k), langLabels.get(k)))
			{
				instanceLists[k].addThruPipe(i);
			}
		}
		
		InstanceList instanceList = new InstanceList(pipe);
		
		Random r = new Random();
		for(int k=0; k<numWordsInSample; ++k)
		{
			for(int j=0; j<instanceLists.length; ++j)
			{
				instanceList.add(instanceLists[j].get(r.nextInt(instanceLists[j].size())));
			}
		}
		
		return instanceList;
	}
	
	public InstanceList readTrainingData(Map<String, File> samplesByID)
	{
		List<String> languageIDs = new ArrayList<String>();
		List<File> languageSampleFiles = new ArrayList<File>();
		for(Map.Entry<String, File> entry : samplesByID.entrySet())
		{
			languageIDs.add(entry.getKey());
			languageSampleFiles.add(entry.getValue());
		}
		return readTrainingData(languageSampleFiles, languageIDs);
	}
	
	public InstanceList readTrainingData(File sampleDir)
	{
		InstanceList instanceList = new InstanceList(pipe);
		Random r = new Random();
		for(File languageSample : sampleDir.listFiles())
		{
			String label = languageSample.getName().substring(0, languageSample.getName().indexOf('.'));
			InstanceList langInstanceList = new InstanceList(pipe);
			for(Instance i : readLanguageTrainingData(languageSample, label))
			{
				langInstanceList.addThruPipe(i);
			}
			for(int k=0; k<numWordsInSample; ++k)
			{
				instanceList.add(langInstanceList.get(r.nextInt(langInstanceList.size())));
			}
		}
		return instanceList;
	}
	
	public Map<String, Map<String, Integer>> readFeatureCounts(File sampleDir)
	{
		Map<String, Map<String, Integer>> featureCountsByLabel = new HashMap<String, Map<String, Integer>>();
		for(File languageSample : sampleDir.listFiles())
		{
			String label = languageSample.getName().substring(0, languageSample.getName().indexOf('.'));
			featureCountsByLabel.put(label, featureCounts(languageSample));
			
		}
		return featureCountsByLabel;
	}
	
	List<Instance> readLanguageTrainingData(File sampleText, String label)
	{
		List<Instance> instances = new ArrayList<Instance>();
		WordInstanceBuilder wordToInstance = new WordInstanceBuilder(features, pipe.getDataAlphabet());
		try
		{
			BufferedReader reader = new BufferedReader(new FileReader(sampleText));
			String line = "";
			while((line = reader.readLine()) != null)
			{
				String[] tokens = line.split("[\\s\\p{Punct}\\d]+");
				for(String token : tokens)
				{
					if(token.length() == 0)
						continue;
					Instance i = wordToInstance.convertWordToInstance(token, label);
					//if(((String)i.getData()).length() > 0)
						instances.add(i);
				}
			}
		}
		catch(IOException ioe)
		{
			ioe.printStackTrace();
		}
		
		return instances;
	}
	
	public Map<String, Integer> featureCounts(File sampleText)
	{
		Map<String, Integer> sampledFeatureCounts = new HashMap<String, Integer>();
		List<List<String>> allFeatureInstances = new ArrayList<List<String>>();
		WordInstanceBuilder wordToInstance = new WordInstanceBuilder(features, pipe.getDataAlphabet());
		try
		{
			BufferedReader reader = new BufferedReader(new FileReader(sampleText));
			String line = "";
			while((line = reader.readLine()) != null)
			{
				String[] tokens = line.split("[\\s\\p{Punct}\\d]+");
				for(String token : tokens)
				{
					if(token.length() == 0)
						continue;
					List<String> wordFeatures = wordToInstance.featureList(token);
					
					allFeatureInstances.add(wordFeatures);
				}
			}
		}
		catch(IOException ioe)
		{
			ioe.printStackTrace();
		}
		
		Random r = new Random();
		for(int k=0; k<numWordsInSample; ++k)
		{
			for(String feature : allFeatureInstances.get(r.nextInt(allFeatureInstances.size())))
			{
				if(sampledFeatureCounts.containsKey(feature))
					sampledFeatureCounts.put(feature, sampledFeatureCounts.get(feature) + 1);
				else
					sampledFeatureCounts.put(feature, 1);
			}
		}
		
		return sampledFeatureCounts;
	}
	
	public Map<String, Integer> featureCountsBinary(File sampleText)
	{
		Map<String, Integer> sampledFeatureCounts = new HashMap<String, Integer>();
		List<List<String>> allFeatureInstances = new ArrayList<List<String>>();
		WordInstanceBuilder wordToInstance = new WordInstanceBuilder(features, pipe.getDataAlphabet());
		try
		{
			BufferedReader reader = new BufferedReader(new FileReader(sampleText));
			String line = "";
			while((line = reader.readLine()) != null)
			{
				String[] tokens = line.split("[\\s\\p{Punct}\\d]+");
				for(String token : tokens)
				{
					if(token.length() == 0)
						continue;
					List<String> wordFeatures = wordToInstance.featureListBinary(token);
					
					allFeatureInstances.add(wordFeatures);
				}
			}
		}
		catch(IOException ioe)
		{
			ioe.printStackTrace();
		}
		
		Random r = new Random();
		for(int k=0; k<numWordsInSample; ++k)
		{
			for(String feature : allFeatureInstances.get(r.nextInt(allFeatureInstances.size())))
			{
				if(sampledFeatureCounts.containsKey(feature))
					sampledFeatureCounts.put(feature, sampledFeatureCounts.get(feature) + 1);
				else
					sampledFeatureCounts.put(feature, 1);
			}
		}
		
		return sampledFeatureCounts;
	}
}
